#!/usr/bin/perl

require "config.pl";
require "$SourceDir/headers.pl";

Preamble();

$pattern="$IncomingDir/*";
@files=glob($pattern);

foreach (@files) {

    #Clean up some variables from previous iterations.
    undef $ResponseMessage;
    undef $status;
    undef $filename;
    %Transaction = ();
   
    $filename=$_;
    if ((!($filename=~/tmp$/))&&(-f $filename)) {
	@parts="";
	@parts=split(/\//,$filename); 
	$numparts=@parts;
	$Filename_without_directory=$parts[$numparts-1];
	$result=&ParseTransactionFile($filename);
	$NumTransactions++;
	close(INFILE);

	if ($result==0) { 
	    $newresult = ActOnTransaction();
	    GenerateTransactionResponse($newresult,$ResponseMessage);

	    #This is for pending transactions who errored out, but can still
	    #be run with the help of an admin.
	    if($newresult > 2000 && $newresult < 3000){ ProcessBadFile($filename); }
	    else{ &ProcessFile($filename) };
	}	    

	else{
	    ProcessBadFile($filename);
	}

    }
}

&Postamble;

exit 0;


### Functions ################################################################
# Read in a transaction file.
# The following checks are done.
# 1) Legitimate transaction type.
# 2) Response or initial transaction.  Associative array -Transaction- built.
# 3) If initial ddtransaction, associative array -Pairs- built of name:value
#    pairs.  Returns error if any fields that are required are undefined or
#    any fields which are neither optional nor required ARE defined.
sub ParseTransactionFile {
    my($filename)=@_[0];
    my($i,$j,$k);
    my($req,$resp,$site,$transaction_type,$transaction_id,$date);
    my($retry,$status,$detail,$message,$transindex);
    my($header_read,$line,$numreq,$numopt,$key,$name,$value);
    my($usage_report);
    my($respfile);
    
    open(INFILE,"$filename")||return -1;
    %Pairs=();

    
    LogMessage("Parsing $Filename_without_directory");

    $header_read=0;
    $usage_report=0;
    while (<INFILE>) {
	$line=$_;
	$line=~s/\n//;

	# Ignore lines that start with the pound (#) symbol as comments.
	if ($line=~/^\#/) { next; }

	# READ HEADER
	# INITIAL TRANSACTION (not a response)
	if ($line=~/^\*transaction/) {
	    ($req,$site_name,$transaction_type,$transaction_id,$date,$retry,
	     @rest)=split(/\s+/,$line);

	    $Transaction{"response"}="";
	    $Transaction{"site"}=$site_name;
	    $Transaction{"type"}=$transaction_type;
	    $Transaction{"id"}=$transaction_id;
	    $Transaction{"index"}=$transindex;
	    $Transaction{"date"}=$date;
	    $Transaction{"retry_count"}=$retry;

	    $transindex=&member_index($transaction_type,@TransactionTypes)-1;

	    # In cases of errors with the header, generate an error rather than
	    # a response.  Let the problem be dealt with by hand.
	    if ($transindex<0) { 
		LogMessage("File: $filename. Unknown transaction type: $transaction_type in transaction RESPONSE \# $transaction_id");
		return 1;
	    }
	    if (($site_name eq "")||($transaction_id eq "")||
		($date eq "")||($retry eq "")) {
		LogMessage("File: $filename.  Wrongly formatted header: $line");
		return 1;
	    }


	    $header_read=1;

	    LogMessage("Header read.");
	    next;
	}
	
	# Ignore everything if you haven't gotten the header line yet.  Assume
	# everything above the header line is comments to be ignored.
	if (!($header_read)) { next; }

	# FOOTER Line.  Deal with and exit.
	if ($line=~/^\*end/) {
	    &LogMessage("Footer read.");

    	    $numreq=$#{$RequiredFields[$transindex]};
    	    $numopt=$#{$OptionalFields[$transindex]};

	    #Make sure all required information is included
	    for($i=0;$i<$numreq;$i++) {
		if ($Pairs{$RequiredFields[$transindex][$i]} eq "") {
		    LogMessage("Required field $RequiredFields[$transindex][$i] undefined.");
		    &GenerateTransactionResponse(3003,"Required field $RequiredFields[$transindex][$i] undefined.");
		    return 1;
		}
	    }

	    foreach (sort (keys %Pairs)) {
		$key=$_;

		# Make sure these are legitimate fields
		if ((!(&member($key,@{$RequiredFields[$transindex]})))
			&& 
		    (!(&member($key,@{$OptionalFields[$transindex]})))) {
		    LogMessage("Field $key is not appropriate for transactions of type $transaction_type");
		    &GenerateTransactionResponse(3004,"Field $key is not appropriate for transactions of type $transaction_type");
		    return 1;
		}
	    }

	    return 0;
	}

	
	#Name:value pair to be parsed
	($name,$value)=split(/:/,$line,2);
	$name=~s/\s+//g; # Spaces not allowed in field names.
	$value=~s/^\s+//; # Remove leading spaces from field values.
	if ($value eq "") { $value="   "; }
	$Pairs{"$name"}="$value";
    }
    if ($header_read) {
	LogMessage("No footer line found in transaction $transaction_id.");
	&GenerateTransactionResponse(3002,"No footer line found in transaction $transaction_id.");
    }
    else {
	if(!defined($transaction_id)){
	    $transaction_id = "file $filename";
	}
	LogMessage("No header line found in transaction $transaction_id.");
       	&GenerateTransactionResponse(3001,"No header line found in transaction $transaction_id.");
    }
    LogMessage("Badly formatted transaction file.");
    return 1;
}


# Generate Response
# Detail Codes:
#  1000-1999: Success
#    Specific success codes.
#    1000: Simple success - file received fine.  No obvious errors.
#    1001: Action previously taken so success by default (for example, a RAC
#          transaction is sent for an account which is already Active).
#  2000-2999: Pending
#  3000-3999: Errors
#    Specific error codes.
#    3001: Header formatted incorrectly.
#    3002: No footer (or misformatted footer).
#    3003: Required name:value pair missing.
#    3004: Name:value pair not allowed for this transaction type.
#    3005: Value is of the wrong format for this field.
#    3006: Format error of some other sort.
#    3100: Inconsistent data (example: partner_psn project doesn't exist)
#    3101: Inappropriate transaction type.

sub GenerateTransactionResponse {
    my($status,$message)=@_;

    if(!defined($Transaction{'id'})){
	$Transaction{'id'} = "UNKNOWN";
    }

    if(!defined($ResponseMessage)){
	$ResponseMessage = $message;
    }
    
    if($status == 1000){
	emailResponse("SUCCESS",$ResponseMessage,$status);
    }
    else{
	emailResponse("FAILED",$ResponseMessage,$status);
    }


}



# Perform some startup commands.
sub Preamble {

    $NumTransactionsCompleted=0;
    $NumTransactions=0;

    if(! (-f "$Logfile")) { `touch $Logfile`; }
    open(LOG, ">>$Logfile") || print "Unable to open logfile $Logfile.\n";

    &LogMessage("Starting.");

    #Make sure our structure is correct.
    if(! (-d "$IncomingDir")) { `mkdir $IncomingDir`; }
    if(! (-d "$IncomingDir/Done")) { `mkdir $IncomingDir/Done`; }
    if(! (-d "$IncomingDir/Errored")) { `mkdir $IncomingDir/Errored`; }
    if(! (-f "$IncomingDir/vmr.data")) { `touch $SourceDir/vmr.data`; }

    #Load the configuration file.
    loadConfig();

    #We have only defined one config file entries, but feel free to add more.
    if(!defined($config{"LastGIBUserId"})) {
	$config{"LastGIBUserId"} = 1;
    }
    close DATAFILE;

    #This is the GiB enabled version, so lets make sure we are root.
    &LogMessage("Grid-in-a-Box Environment Enabled.");
    if($< != 0) {
	&LogMessage("Not running as root! 'adduser' will break! Exiting.");
        #we want to exit prematurely because we dont want to write
        #out anything new to the config files, this maintains state.
        exit 1;
    }
    $date = `date '+%b %-d %H:%M:%S'`;
    chomp $date;
    $tempBuffer = "/tmp/vmr.tmp.$$";
    unless(open (TB, ">$tempBuffer")){
	LogMessage("Unable to create temporary file $tempBuffer. $!");
    }
    
    print TB "Started at $date.\n";
    
}

# Close log file descriptors and run the MD5 check facility.
sub Postamble {
    local $file;
    
    if($NumTransactionsCompleted != 0 && $MD5Check =~ /rue/) {
	checkFileConsistency();    
    }
    
    writeConfig();
    
    &LogMessage("Executed $NumTransactions transactions, $NumTransactionsCompleted successfull. Exiting.");
    
    close TB;
    if($NumTransactions != 0){
	unless(open (MAIL, "|/bin/mail -s \"GiB Account System \[ $SiteDomain \] \" $EmailContact")){
	    LogMessage("Could not open mail buffer to email output. $!");
	}
	unless(open (TB, "$tempBuffer")){
	    LogMessage("Could not open temporary file $tempBuffer to email response. $!");
	}
	while(<TB>){
	    print MAIL;
	}
	
	close TB;
	close MAIL;
    }
    
    `rm $tempBuffer`;
    close LOG;
}

sub checkFileConsistency{

    if(! (-d "$SourceDir/tmp")) { `mkdir $SourceDir/tmp`; }

    #Here we are going to run md5 checksums on all the passwd files and
    #make sure that they are the same, if not, send them over again.
    
    #We assume the file locations are the same between cluster machines.

    #This is the list of files we want to get from the machines on the 
    #cluster. If there are more, add them.
    @files = ('/etc/passwd','/etc/group','/etc/shadow',"/etc/gshadow");

    #For each of these files on our current system, we load the file name
    #as the key and its md5 equivalent as its value.
    %newHash = LoadMD5(@files);

    #We attempt to get each of these files.
    foreach(@files) {
	$rc = system("$CGET -s $_ -t $SourceDir/tmp/");
		
	if($rc != 0) {
	    $rc = $rc/256;
	    LogMessage("$CGET returned $rc. Couldnt update passwd db.");
	    exit 1; 
	}
	push @newfiles, glob("$SourceDir/tmp/*");
    }
    
    
    #Get downloaded files md5 sums.
    %realHash = LoadMD5(@newfiles);

    foreach $key (keys %newHash) {
	foreach $subval (@{ $realHash{$key} }){
	    if($newHash{$key} ne $subval){
		$rc = system("$SYNC");
		if($rc != 0) {
		    $rc = $rc/256;
		    LogMessage("$SYNC returned $rc. Couldnt update passwd db.");
		    exit 1;
		}
		print "$newHash{$key} != $subval\n";
		last;
	    }
	}
    }
	
    `rm -rf $SourceDir/tmp/`;

}


sub writeConfig{
#Write out the config file.
    unless (open(DATAFILE, ">$SourceDir/vmr.data")){
	LogMessage("Unable to rewrite datafile: $!");
    }

    foreach $key (keys %config){
	print DATAFILE "$key:".$config{$key}."\n";
    }
    close DATAFILE;

}


#This is where we load the VMR data file.
sub loadConfig{
    %config = ();
    unless (open(DATAFILE,"$SourceDir/vmr.data")) {
	LogMessage("vmr.data could not be read. $!");        
    }
    while(<DATAFILE>) {
	$line = $_;
        chomp($line);
        if($line=~/^\#/) { next; }
	($name,$value)=split(/:/,$line,2);
        if($value eq ""){
	    $value="0";
	}
        $config{$name}=$value;
    }

}


sub LoadMD5 {
    local (@files = @_, $file, %md5hash = ());

    foreach(@files) {
	@parts=split(/\//,$_);
        $numparts=@parts;
        $file=$parts[$numparts-1];
	$hash = `$MD5SUM $_`;
	chomp($hash);
	$hash =~ /(\w{32})/;
	$hash = $1;
	
	if($file =~ /_/) {
	    ($file, $crap) = split(/_/,$file);
	}

	#We make a hash of arrays, so that we can maintain several
	#md5sums for individual files. If one of them differs, we
	#try and syncronize the user db again.

	if(defined($md5hash{$file})){
	    push @{ $md5hash{$file} }, $hash; 
	}

	$md5hash{$file} = ( $hash );
    }

    foreach $key (keys %md5hash) {
	print "$key = $md5hash{$key}\n";
    }

    return %md5hash;

}

# We have received an Incoming Transaction.  Do something about it.
# Available associative arrays to this function and others in this file:
#   Pairs        :: composed of name:value pairs
#   Transaction  :: information about the transaction itself 
# Return positive DETAIL_CODE to ask for a return of an appropriate Failure.
# $ResponseMessage should also be set to something appropriate.

sub ActOnTransaction {
    
    if($Transaction{"type"} eq "REQUEST_ACCOUNT_CREATE") {
	$status = &CreateAccount;
    } 
    if($status == 1000){
        return $status;
    }
   
    return $status; #error
}



# Move Transaction file which has already been dealt with to the appropriate place.
sub ProcessFile {
    my($filename)=$_[0];

    @parts=split(/\//,$filename);
    $numparts=@parts;
    $filenodir=$parts[$numparts-1];

    $rc = system "mv $filename $IncomingDir/Done/$filenodir";
    if($rc != 0){
	LogMessage("Error moving $filename to $IncomingDir/Done/");
    }
}

#Move transaction file to the errored directory..
sub ProcessBadFile {
    my($filename)=$_[0];

    @parts=split(/\//,$filename);
    $numparts=@parts;
    $filenodir=$parts[$numparts-1];

    $rc = system "mv $filename $IncomingDir/Errored/$filenodir";
    if($rc != 0){
	LogMessage("Error moving $filename to $IncomingDir/Errored/");
    }
}




#REQUEST_ACCOUNT_CREATE w/ GIB.
#Create an account immediatly using the GIB infrastructure.
#Runs the adduser program locally which syncs the cluster machines passwd db with the local db.
#It also writes the DN to the gridmapfile to $Gridfile
sub CreateAccount {
    my($username);

    unless (open(GRIDMAPFILE,"$Gridfile")){
	$ResponseMessage = "Could not read gridmap file $Gridfile. $!";
	LogMessage($ResponseMessage);
	return 2500;
    }
 
   while(<GRIDMAPFILE>){
	$line = $_;
	chomp($line);
	if($line =~ /$Pairs{'dn'}/){
	    $ResponseMessage = "DN \"$Pairs{'dn'}\" already in grid file $Gridfile.";
	    LogMessage($ResponseMessage);
	    return 2500;
	}
    }

    unless (open(GRIDMAPFILE,">>$Gridfile")) {
	$ResponseMessage = "Could not write to grid map file $Gridfile. $!";
	LogMessage($ResponseMessage);
	return 2500;
    }

    $mygoodid = sprintf("%03u", $config{"LastGIBUserId"});
    $username = "gib-".$mygoodid;
    $config{"LastGIBUserId"} = $config{"LastGIBUserId"} + 1;
    
    print GRIDMAPFILE "\"".$Pairs{"dn"}."\""." "."$username\n";
    close GRIDMAPFILE;

    $rc = system("/usr/sbin/adduser $username");

    if($rc != 0) {
	$rc = $rc/256;
	$ResponseMessage = "User not created, \"adduser\" returned $rc. $!";
	LogMessage($ResponseMessage);
	return 2500;
    }

    $ResponseMessage = "Account $username created successfully.";
    LogMessage($ResponseMessage);
    $NumTransactionsCompleted++;

    return 1000;
}


#This will send an email to the designated administrator with notification that the account 
#was created successfully or unsuccessfully.

sub emailResponse{

    my ($state,$ResponseMessage,$status)=@_;
    if($state eq "SUCCESS"){ 
	$state = "successfully"; 
    }
    else { 
	$state = "unsuccessfully"; 
    }
    
    print TB "------------------------------------------------------\n";
    print TB "Transaction $Transaction{'id'} executed $state.\n";
    print TB "\n";
    print TB "Response: $ResponseMessage\n";
    print TB "Status: $status\n";
    if(!defined($Pairs{'dn'})){
	print TB "DN: unknown\n";
    }
    else{
	print TB "DN: $Pairs{'dn'}\n";
    }

    if($state eq "unsuccessfully"){
	print TB "\n";
	print TB "Notify the site administrator of this error by having him or her\n";
	print TB "examine the error directory and the logfile.\n";
    }
    

}


# SEE IF AN ITEM IS PART OF AN ARRAY
sub member {
    my($matchitem,@myarray)=@_;
    my($retval)=0;
    my($item);

    foreach $item (@myarray) {
        if ("$matchitem" eq "$item") {
            $retval=1;
        }
    }

    $retval;
}


# SEE IF AN ITEM IS PART OF AN ARRAY and return its index value + 1
sub member_index {
    my($matchitem,@myarray)=@_;
    my($i,$numitems);
    my($retval)=0;

    $numitems=@myarray;
    for($i=0;$i<$numitems;$i++) {
        if ("$matchitem" eq "$myarray[$i]") {
            $retval = $i+1;
        }
    }

    $retval;
}


#If this breaks on your machine its probably because your version of
#'date' has different switches than the one from the sh-utils package.        
sub LogMessage {
    my($logmessage)=$_[0];
    
    $date = `date '+%b %-d %H:%M:%S'`;
    chomp($date);
    print "$date: $logmessage\n";
    print LOG "$date: $logmessage\n";
}


1;
